@@ -38,8 +38,8 @@
    19990319 - Arnaldo Carvalho de Melo <acme@conectiva.com.br> - i18n/nls
 
    20000101 - David Huggins-Daines <dhuggins@linuxcare.com> - Better
-   support for OSF/1 disklabels on Alpha.  Also fixed unaligned
-   accesses in alpha_bootblock_checksum()
+   support for OSF/1 disklabels on Alpha.
+   Also fixed unaligned accesses in alpha_bootblock_checksum()
 */
 
 #include <unistd.h>
@@ -52,14 +52,12 @@
 #include <errno.h>
 #include "nls.h"
 
-#include <sys/ioctl.h>
 #include <sys/param.h>
 
-#include <linux/hdreg.h>	/* for HDIO_GETGEO */
-
 #include "common.h"
 #include "fdisk.h"
-#define NETBSD_PARTITION 0xa5
+#define FREEBSD_PARTITION	0xa5
+#define NETBSD_PARTITION	0xa9
 #define DKTYPENAMES
 #include "fdiskbsdlabel.h"
 
@@ -73,8 +71,9 @@ static void xbsd_change_fstype (void);
 static int xbsd_get_part_index (int max);
 static int xbsd_check_new_partition (int *i);
 static void xbsd_list_types (void);
-static u_short xbsd_dkcksum (struct xbsd_disklabel *lp);
-static int xbsd_initlabel  (struct partition *p, struct xbsd_disklabel *d, int pindex);
+static unsigned short xbsd_dkcksum (struct xbsd_disklabel *lp);
+static int xbsd_initlabel  (struct partition *p, struct xbsd_disklabel *d,
+			    int pindex);
 static int xbsd_readlabel  (struct partition *p, struct xbsd_disklabel *d);
 static int xbsd_writelabel (struct partition *p, struct xbsd_disklabel *d);
 static void sync_disks (void);
@@ -102,21 +101,26 @@ static struct xbsd_disklabel xbsd_dlabel;
 #define bsd_cround(n) \
 	(display_in_cyl_units ? ((n)/xbsd_dlabel.d_secpercyl) + 1 : (n))
 
+/*
+ * Test whether the whole disk has BSD disk label magic.
+ *
+ * Note: often reformatting with DOS-type label leaves the BSD magic,
+ * so this does not mean that there is a BSD disk label.
+ */
 int
 check_osf_label(void) {
-    if (xbsd_readlabel (NULL, &xbsd_dlabel) == 0)
-	return 0;
-    osf_label = 1;
-    return 1;
+	if (xbsd_readlabel (NULL, &xbsd_dlabel) == 0)
+		return 0;
+	return 1;
 }
 
 int
 btrydev (char * dev) {
-    if (xbsd_readlabel (NULL, &xbsd_dlabel) == 0)
-	return -1;
-    printf(_("\nBSD label for device: %s\n"), dev);
-    xbsd_print_disklabel (0);
-    return 0;
+	if (xbsd_readlabel (NULL, &xbsd_dlabel) == 0)
+		return -1;
+	printf(_("\nBSD label for device: %s\n"), dev);
+	xbsd_print_disklabel (0);
+	return 0;
 }
 
 static void
@@ -147,8 +151,11 @@ hidden(int type) {
 }
 
 static int
-is_netbsd_partition_type(int type) {
-	return (type == NETBSD_PARTITION || type == hidden(NETBSD_PARTITION));
+is_bsd_partition_type(int type) {
+	return (type == FREEBSD_PARTITION ||
+		type == hidden(FREEBSD_PARTITION) ||
+		type == NETBSD_PARTITION ||
+		type == hidden(NETBSD_PARTITION));
 }
 #endif
 
@@ -160,7 +167,7 @@ bselect (void) {
 
   for (t=0; t<4; t++) {
     p = get_part_table(t);
-    if (p && is_netbsd_partition_type(p->sys_ind)) {
+    if (p && is_bsd_partition_type(p->sys_ind)) {
       xbsd_part = p;
       xbsd_part_index = t;
       ss = get_start_sect(xbsd_part);
@@ -258,17 +265,17 @@ xbsd_delete_part (void)
 static void
 xbsd_new_part (void)
 {
-  uint begin, end;
+  unsigned int begin, end;
   char mesg[256];
   int i;
 
   if (!xbsd_check_new_partition (&i))
     return;
 
-#if !defined (__alpha__) && !defined (__powerpc__)
+#if !defined (__alpha__) && !defined (__powerpc__) && !defined (__hppa__)
   begin = get_start_sect(xbsd_part);
   end = begin + get_nr_sects(xbsd_part) - 1;
-#elif defined (__alpha__) || defined (__powerpc__)
+#else
   begin = 0;
   end = xbsd_dlabel.d_secperunit - 1;
 #endif
@@ -419,7 +426,8 @@ xbsd_create_disklabel (void) {
 		c = read_char (_("Do you want to create a disklabel? (y/n) "));
 		if (tolower(c) == 'y') {
 			if (xbsd_initlabel (
-#if defined (__alpha__) || defined (__powerpc__)
+#if defined (__alpha__) || defined (__powerpc__) || defined (__hppa__) || \
+    defined (__s390__) || defined (__s390x__)
 				NULL, &xbsd_dlabel, 0
 #else
 				xbsd_part, &xbsd_dlabel, xbsd_part_index
@@ -440,7 +448,7 @@ edit_int (int def, char *mesg)
   do {
     fputs (mesg, stdout);
     printf (" (%d): ", def);
-    if (!read_line ())
+    if (!read_line (NULL))
       return def;
   }
   while (!isdigit (*line_ptr));
@@ -455,28 +463,28 @@ xbsd_edit_disklabel (void)
   d = &xbsd_dlabel;
 
 #if defined (__alpha__) || defined (__ia64__)
-  d -> d_secsize    = (u_long) edit_int ((u_long) d -> d_secsize     ,_("bytes/sector"));
-  d -> d_nsectors   = (u_long) edit_int ((u_long) d -> d_nsectors    ,_("sectors/track"));
-  d -> d_ntracks    = (u_long) edit_int ((u_long) d -> d_ntracks     ,_("tracks/cylinder"));
-  d -> d_ncylinders = (u_long) edit_int ((u_long) d -> d_ncylinders  ,_("cylinders"));
+  d -> d_secsize    = (unsigned long) edit_int ((unsigned long) d -> d_secsize     ,_("bytes/sector"));
+  d -> d_nsectors   = (unsigned long) edit_int ((unsigned long) d -> d_nsectors    ,_("sectors/track"));
+  d -> d_ntracks    = (unsigned long) edit_int ((unsigned long) d -> d_ntracks     ,_("tracks/cylinder"));
+  d -> d_ncylinders = (unsigned long) edit_int ((unsigned long) d -> d_ncylinders  ,_("cylinders"));
 #endif
 
   /* d -> d_secpercyl can be != d -> d_nsectors * d -> d_ntracks */
   while (1)
   {
-    d -> d_secpercyl = (u_long) edit_int ((u_long) d -> d_nsectors * d -> d_ntracks,
+    d -> d_secpercyl = (unsigned long) edit_int ((unsigned long) d -> d_nsectors * d -> d_ntracks,
 					  _("sectors/cylinder"));
     if (d -> d_secpercyl <= d -> d_nsectors * d -> d_ntracks)
       break;
 
     printf (_("Must be <= sectors/track * tracks/cylinder (default).\n"));
   }
-  d -> d_rpm        = (u_short) edit_int ((u_short) d -> d_rpm       ,_("rpm"));
-  d -> d_interleave = (u_short) edit_int ((u_short) d -> d_interleave,_("interleave"));
-  d -> d_trackskew  = (u_short) edit_int ((u_short) d -> d_trackskew ,_("trackskew"));
-  d -> d_cylskew    = (u_short) edit_int ((u_short) d -> d_cylskew   ,_("cylinderskew"));
-  d -> d_headswitch = (u_long) edit_int ((u_long) d -> d_headswitch  ,_("headswitch"));
-  d -> d_trkseek    = (u_long) edit_int ((u_long) d -> d_trkseek     ,_("track-to-track seek"));
+  d -> d_rpm        = (unsigned short) edit_int ((unsigned short) d -> d_rpm       ,_("rpm"));
+  d -> d_interleave = (unsigned short) edit_int ((unsigned short) d -> d_interleave,_("interleave"));
+  d -> d_trackskew  = (unsigned short) edit_int ((unsigned short) d -> d_trackskew ,_("trackskew"));
+  d -> d_cylskew    = (unsigned short) edit_int ((unsigned short) d -> d_cylskew   ,_("cylinderskew"));
+  d -> d_headswitch = (unsigned long) edit_int ((unsigned long) d -> d_headswitch  ,_("headswitch"));
+  d -> d_trkseek    = (unsigned long) edit_int ((unsigned long) d -> d_trkseek     ,_("track-to-track seek"));
 
   d -> d_secperunit = d -> d_secpercyl * d -> d_ncylinders;
 }
@@ -506,7 +514,7 @@ static void
 xbsd_write_bootstrap (void)
 {
   char *bootdir = BSD_LINUX_BOOTDIR;
-  char path[MAXPATHLEN];
+  char path[sizeof(BSD_LINUX_BOOTDIR) + 1 + 2 + 4];  /* BSD_LINUX_BOOTDIR + / + {sd,wd} + boot */
   char *dkbasename;
   struct xbsd_disklabel dl;
   char *d, *p, *e;
@@ -519,7 +527,7 @@ xbsd_write_bootstrap (void)
 
   printf (_("Bootstrap: %sboot -> boot%s (%s): "),
 	  dkbasename, dkbasename, dkbasename);
-  if (read_line ()) {
+  if (read_line (NULL)) {
     line_ptr[strlen (line_ptr)-1] = '\0';
     dkbasename = line_ptr;
   }
@@ -529,10 +537,10 @@ xbsd_write_bootstrap (void)
 
   /* We need a backup of the disklabel (xbsd_dlabel might have changed). */
   d = &disklabelbuffer[BSD_LABELSECTOR * SECTOR_SIZE];
-  bcopy (d, &dl, sizeof (struct xbsd_disklabel));
+  memmove (&dl, d, sizeof (struct xbsd_disklabel));
 
   /* The disklabel will be overwritten by 0's from bootxx anyway */
-  bzero (d, sizeof (struct xbsd_disklabel));
+  memset (d, 0, sizeof (struct xbsd_disklabel));
 
   snprintf (path, sizeof(path), "%s/boot%s", bootdir, dkbasename);
   if (!xbsd_get_bootstrap (path, &disklabelbuffer[xbsd_dlabel.d_secsize],
@@ -546,9 +554,9 @@ xbsd_write_bootstrap (void)
       exit ( EXIT_FAILURE );
     }
 
-  bcopy (&dl, d, sizeof (struct xbsd_disklabel));
+  memmove (d, &dl, sizeof (struct xbsd_disklabel));
 
-#if defined (__powerpc__)
+#if defined (__powerpc__) || defined (__hppa__)
   sector = 0;
 #elif defined (__alpha__)
   sector = 0;
@@ -557,7 +565,7 @@ xbsd_write_bootstrap (void)
   sector = get_start_sect(xbsd_part);
 #endif
 
-  if (ext2_llseek (fd, (ext2_loff_t) sector * SECTOR_SIZE, SEEK_SET) == -1)
+  if (lseek (fd, (off_t) sector * SECTOR_SIZE, SEEK_SET) == -1)
     fatal (unable_to_seek);
   if (BSD_BBSIZE != write (fd, disklabelbuffer, BSD_BBSIZE))
     fatal (unable_to_write);
@@ -595,162 +603,171 @@ xbsd_get_part_index (int max)
 }
 
 static int
-xbsd_check_new_partition (int *i)
-{
-  int t;
+xbsd_check_new_partition (int *i) {
 
-  if (xbsd_dlabel.d_npartitions == BSD_MAXPARTITIONS)
-  {
-    for (t=0; t < BSD_MAXPARTITIONS; t++)
-      if (xbsd_dlabel.d_partitions[t].p_size == 0)
-	break;
+	/* room for more? various BSD flavours have different maxima */
+	if (xbsd_dlabel.d_npartitions == BSD_MAXPARTITIONS) {
+		int t;
 
-    if (t == BSD_MAXPARTITIONS)
-    {
-      fprintf (stderr, _("The maximum number of partitions has been created\n"));
-      return 0;
-    }
-  }
-  *i = xbsd_get_part_index (BSD_MAXPARTITIONS);
+		for (t = 0; t < BSD_MAXPARTITIONS; t++)
+			if (xbsd_dlabel.d_partitions[t].p_size == 0)
+				break;
 
-  if (*i >= xbsd_dlabel.d_npartitions)
-    xbsd_dlabel.d_npartitions = (*i) + 1;
+		if (t == BSD_MAXPARTITIONS) {
+			fprintf (stderr, _("The maximum number of partitions "
+					   "has been created\n"));
+			return 0;
+		}
+	}
 
-  if (xbsd_dlabel.d_partitions[*i].p_size != 0)
-  {
-    fprintf (stderr, _("This partition already exists.\n"));
-    return 0;
-  }
-  return 1;
+	*i = xbsd_get_part_index (BSD_MAXPARTITIONS);
+
+	if (*i >= xbsd_dlabel.d_npartitions)
+		xbsd_dlabel.d_npartitions = (*i) + 1;
+
+	if (xbsd_dlabel.d_partitions[*i].p_size != 0) {
+		fprintf (stderr, _("This partition already exists.\n"));
+		return 0;
+	}
+
+	return 1;
 }
 
 static void
-xbsd_list_types (void)
-{
-  list_types (xbsd_fstypes);
+xbsd_list_types (void) {
+	list_types (xbsd_fstypes);
 }
 
-static u_short
-xbsd_dkcksum (struct xbsd_disklabel *lp)
-{
-  register u_short *start, *end;
-  register u_short sum = 0;
+static unsigned short
+xbsd_dkcksum (struct xbsd_disklabel *lp) {
+	unsigned short *start, *end;
+	unsigned short sum = 0;
   
-  start = (u_short *)lp;
-  end = (u_short *)&lp->d_partitions[lp->d_npartitions];
-  while (start < end)
-    sum ^= *start++;
-  return (sum);
+	start = (unsigned short *) lp;
+	end = (unsigned short *) &lp->d_partitions[lp->d_npartitions];
+	while (start < end)
+		sum ^= *start++;
+	return sum;
 }
 
 static int
-xbsd_initlabel (struct partition *p, struct xbsd_disklabel *d, int pindex)
-{
-  struct xbsd_partition *pp;
-  struct geom g;
+xbsd_initlabel (struct partition *p, struct xbsd_disklabel *d,
+		int pindex __attribute__((__unused__))) {
+	struct xbsd_partition *pp;
+	struct geom g;
 
-  get_geometry (fd, &g);
-  bzero (d, sizeof (struct xbsd_disklabel));
+	get_geometry (fd, &g);
+	memset (d, 0, sizeof (struct xbsd_disklabel));
 
-  d -> d_magic = BSD_DISKMAGIC;
+	d -> d_magic = BSD_DISKMAGIC;
 
-  if (strncmp (disk_device, "/dev/sd", 7) == 0)
-    d -> d_type = BSD_DTYPE_SCSI;
-  else
-    d -> d_type = BSD_DTYPE_ST506;
+	if (strncmp (disk_device, "/dev/sd", 7) == 0)
+		d -> d_type = BSD_DTYPE_SCSI;
+	else
+		d -> d_type = BSD_DTYPE_ST506;
 
 #if 0 /* not used (at least not written to disk) by NetBSD/i386 1.0 */
-  d -> d_subtype = BSD_DSTYPE_INDOSPART & pindex;
+	d -> d_subtype = BSD_DSTYPE_INDOSPART & pindex;
 #endif
 
 #if !defined (__alpha__)
-  d -> d_flags = BSD_D_DOSPART;
+	d -> d_flags = BSD_D_DOSPART;
 #else
-  d -> d_flags = 0;
+	d -> d_flags = 0;
 #endif
-  d -> d_secsize = SECTOR_SIZE;			/* bytes/sector  */
-  d -> d_nsectors = g.sectors;			/* sectors/track */
-  d -> d_ntracks = g.heads;			/* tracks/cylinder (heads) */
-  d -> d_ncylinders = g.cylinders;
-  d -> d_secpercyl  = g.sectors * g.heads;	/* sectors/cylinder */
-  if (d -> d_secpercyl == 0)
-	  d -> d_secpercyl = 1;			/* avoid segfaults */
-  d -> d_secperunit = d -> d_secpercyl * d -> d_ncylinders;
+	d -> d_secsize = SECTOR_SIZE;		/* bytes/sector  */
+	d -> d_nsectors = g.sectors;		/* sectors/track */
+	d -> d_ntracks = g.heads;		/* tracks/cylinder (heads) */
+	d -> d_ncylinders = g.cylinders;
+	d -> d_secpercyl  = g.sectors * g.heads;/* sectors/cylinder */
+	if (d -> d_secpercyl == 0)
+		d -> d_secpercyl = 1;		/* avoid segfaults */
+	d -> d_secperunit = d -> d_secpercyl * d -> d_ncylinders;
 
-  d -> d_rpm = 3600;
-  d -> d_interleave = 1;
-  d -> d_trackskew = 0;
-  d -> d_cylskew = 0;
-  d -> d_headswitch = 0;
-  d -> d_trkseek = 0;
+	d -> d_rpm = 3600;
+	d -> d_interleave = 1;
+	d -> d_trackskew = 0;
+	d -> d_cylskew = 0;
+	d -> d_headswitch = 0;
+	d -> d_trkseek = 0;
 
-  d -> d_magic2 = BSD_DISKMAGIC;
-  d -> d_bbsize = BSD_BBSIZE;
-  d -> d_sbsize = BSD_SBSIZE;
+	d -> d_magic2 = BSD_DISKMAGIC;
+	d -> d_bbsize = BSD_BBSIZE;
+	d -> d_sbsize = BSD_SBSIZE;
 
 #if !defined (__alpha__)
-  d -> d_npartitions = 4;
-  pp = &d -> d_partitions[2]; /* Partition C should be the NetBSD partition */
-  pp -> p_offset = get_start_sect(p);
-  pp -> p_size   = get_nr_sects(p);
-  pp -> p_fstype = BSD_FS_UNUSED;
-  pp = &d -> d_partitions[3]; /* Partition D should be the whole disk */
-  pp -> p_offset = 0;
-  pp -> p_size   = d -> d_secperunit;
-  pp -> p_fstype = BSD_FS_UNUSED;
+	d -> d_npartitions = 4;
+	pp = &d -> d_partitions[2];		/* Partition C should be
+						   the NetBSD partition */
+	pp -> p_offset = get_start_sect(p);
+	pp -> p_size   = get_nr_sects(p);
+	pp -> p_fstype = BSD_FS_UNUSED;
+	pp = &d -> d_partitions[3];		/* Partition D should be
+						   the whole disk */
+	pp -> p_offset = 0;
+	pp -> p_size   = d -> d_secperunit;
+	pp -> p_fstype = BSD_FS_UNUSED;
 #elif defined (__alpha__)
-  d -> d_npartitions = 3;
-  pp = &d -> d_partitions[2]; /* Partition C should be the whole disk */
-  pp -> p_offset = 0;
-  pp -> p_size   = d -> d_secperunit;
-  pp -> p_fstype = BSD_FS_UNUSED;  
+	d -> d_npartitions = 3;
+	pp = &d -> d_partitions[2];		/* Partition C should be
+						   the whole disk */
+	pp -> p_offset = 0;
+	pp -> p_size   = d -> d_secperunit;
+	pp -> p_fstype = BSD_FS_UNUSED;  
 #endif
 
-  return 1;
+	return 1;
 }
 
+/*
+ * Read a xbsd_disklabel from sector 0 or from the starting sector of p.
+ * If it has the right magic, return 1.
+ */
 static int
 xbsd_readlabel (struct partition *p, struct xbsd_disklabel *d)
 {
-  int t, sector;
+	int t, sector;
 
+	/* p is used only to get the starting sector */
 #if !defined (__alpha__)
-  sector = (p ? get_start_sect(p) : 0);
+	sector = (p ? get_start_sect(p) : 0);
 #elif defined (__alpha__)
-  sector = 0;
+	sector = 0;
 #endif
 
-  if (ext2_llseek (fd, (ext2_loff_t) sector * SECTOR_SIZE, SEEK_SET) == -1)
-    fatal (unable_to_seek);
-  if (BSD_BBSIZE != read (fd, disklabelbuffer, BSD_BBSIZE))
-    fatal (unable_to_read);
+	if (lseek (fd, (off_t) sector * SECTOR_SIZE, SEEK_SET) == -1)
+		fatal (unable_to_seek);
+	if (BSD_BBSIZE != read (fd, disklabelbuffer, BSD_BBSIZE))
+		fatal (unable_to_read);
 
-  bcopy (&disklabelbuffer[BSD_LABELSECTOR * SECTOR_SIZE + BSD_LABELOFFSET],
-	 d, sizeof (struct xbsd_disklabel));
+	memmove (d,
+	         &disklabelbuffer[BSD_LABELSECTOR * SECTOR_SIZE + BSD_LABELOFFSET],
+	         sizeof (struct xbsd_disklabel));
 
-  for (t = d -> d_npartitions; t < BSD_MAXPARTITIONS; t++)
-  {
-    d -> d_partitions[t].p_size   = 0;
-    d -> d_partitions[t].p_offset = 0;
-    d -> d_partitions[t].p_fstype = BSD_FS_UNUSED;  
-  }
-  if (d -> d_magic != BSD_DISKMAGIC || d -> d_magic2 != BSD_DISKMAGIC)
-    return 0;
+	if (d -> d_magic != BSD_DISKMAGIC || d -> d_magic2 != BSD_DISKMAGIC)
+		return 0;
 
-  if (d -> d_npartitions > BSD_MAXPARTITIONS)
-    fprintf (stderr, _("Warning: too many partitions (%d, maximum is %d).\n"),
-	     d -> d_npartitions, BSD_MAXPARTITIONS);
-  return 1;
+	for (t = d -> d_npartitions; t < BSD_MAXPARTITIONS; t++) {
+		d -> d_partitions[t].p_size   = 0;
+		d -> d_partitions[t].p_offset = 0;
+		d -> d_partitions[t].p_fstype = BSD_FS_UNUSED;  
+	}
+
+	if (d -> d_npartitions > BSD_MAXPARTITIONS)
+		fprintf (stderr, _("Warning: too many partitions "
+				   "(%d, maximum is %d).\n"),
+			 d -> d_npartitions, BSD_MAXPARTITIONS);
+	return 1;
 }
 
 static int
 xbsd_writelabel (struct partition *p, struct xbsd_disklabel *d)
 {
-  int sector;
+  unsigned int sector;
 
-#if !defined (__alpha__) && !defined (__powerpc__)
+#if !defined (__alpha__) && !defined (__powerpc__) && !defined (__hppa__)
   sector = get_start_sect(p) + BSD_LABELSECTOR;
-#elif defined (__alpha__) || defined (__powerpc__)
+#else
   sector = BSD_LABELSECTOR;
 #endif
 
@@ -760,17 +777,17 @@ xbsd_writelabel (struct partition *p, struct xbsd_disklabel *d)
   /* This is necessary if we want to write the bootstrap later,
      otherwise we'd write the old disklabel with the bootstrap.
   */
-  bcopy (d, &disklabelbuffer[BSD_LABELSECTOR * SECTOR_SIZE + BSD_LABELOFFSET],
-	 sizeof (struct xbsd_disklabel));
+  memmove (&disklabelbuffer[BSD_LABELSECTOR * SECTOR_SIZE + BSD_LABELOFFSET], d,
+           sizeof (struct xbsd_disklabel));
 
 #if defined (__alpha__) && BSD_LABELSECTOR == 0
   alpha_bootblock_checksum (disklabelbuffer);
-  if (ext2_llseek (fd, (ext2_loff_t) 0, SEEK_SET) == -1)
+  if (lseek (fd, (off_t) 0, SEEK_SET) == -1)
     fatal (unable_to_seek);
   if (BSD_BBSIZE != write (fd, disklabelbuffer, BSD_BBSIZE))
     fatal (unable_to_write);
 #else
-  if (ext2_llseek (fd, (ext2_loff_t) sector * SECTOR_SIZE + BSD_LABELOFFSET,
+  if (lseek (fd, (off_t) sector * SECTOR_SIZE + BSD_LABELOFFSET,
 		   SEEK_SET) == -1)
     fatal (unable_to_seek);
   if (sizeof (struct xbsd_disklabel) != write (fd, d, sizeof (struct xbsd_disklabel)))
